# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2023 Stefan Bender
#
# This module is part of pymipas_noxy.
# pymipas_noxy is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying LICENSE file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Correlation histogram utility functions

Utility functions to correlate trace gases against each other.
Initially designed to correlate NOy and CH4 to separate
"background" from "EPP-induced" NOy.
"""
from __future__ import absolute_import, division, print_function

from logging import info, debug

import numpy as np
import xarray as xr

from scipy.stats import gaussian_kde


# %%
def histogram2d_colwise(ds, x, y, x_edges, y_edges, density=False):
    """2-D Histogram with column-wise 1-D histograms

    Assembles a 2-D histogram of the data by grouping "column-wise"
    1-D histograms.

    Parameters
    ----------
    ds: xarray.Dataset
        Input dataset.
    x: str
        Name of the x-variable in the dataset.
    y: str
        Name of the y-variable in the dataset.
    x_edges: array-like (M,)
        Bin edges of the x-variable.
    y_edges: array-like (N,)
        Bin edges of the y-variable.
    density: bool, optional (default False)
        Whether or not to normalize the individual histograms
        to unit density.

    Returns
    -------
    ret: xarray.Dataset
    """
    x_ctr = 0.5 * (x_edges[:-1] + x_edges[1:])
    y_ctr = 0.5 * (y_edges[:-1] + y_edges[1:])
    y_binv = y + "_bins"
    ret = ds.groupby_bins(x, x_edges, labels=x_ctr).apply(
        lambda d: xr.DataArray(
            np.histogram(d[y], bins=y_edges, density=density)[0],
            dims=(y_binv,),
            coords={y_binv: y_ctr},
        )
    )
    return ret


# %%
def histogram2d(ds, x, y, x_edges, y_edges, density=False):
    """2-D Histogram with `numpy.histogram2d`

    Produces a 2-D histogram using numpy's `histogram2d()` function.

    Parameters
    ----------
    ds: xarray.Dataset
        Input dataset.
    x: str
        Name of the x-variable in the dataset.
    y: str
        Name of the y-variable in the dataset.
    x_edges: array-like (M,)
        Bin edges of the x-variable.
    y_edges: array-like (N,)
        Bin edges of the y-variable.
    density: bool, optional (default False)
        Whether or not to normalize the 2-D histogram to unit density.

    Returns
    -------
    ret: xarray.DataArray
        DataArray containing the histogram with dimensions and
        coordinates set accordingly.

    See Also
    --------
    numpy.histogram2d
    """
    x_ctr = 0.5 * (x_edges[:-1] + x_edges[1:])
    y_ctr = 0.5 * (y_edges[:-1] + y_edges[1:])
    x_binv = x + "_bins"
    y_binv = y + "_bins"

    _hist = np.histogram2d(
        ds[x].values.ravel(),
        ds[y].values.ravel(),
        bins=[x_edges, y_edges],
        density=density,
    )
    ret = xr.DataArray(
        _hist[0],
        dims=(x_binv, y_binv,),
        coords={x_binv: x_ctr, y_binv: y_ctr}
    )
    return ret


# %%
def histogram2d_kde(
    ds, x, y, x_edges, y_edges,
    dims=("altitude", "geo_id",),
    **kwargs
):
    """2-D Histogram using Gaussian kernel density estimate

    Calculates a 2-D histogram using scipy's Gaussian kernel density.

    Parameters
    ----------
    ds: xarray.Dataset
        Input dataset.
    x: str
        Name of the x-variable in the dataset.
    y: str
        Name of the y-variable in the dataset.
    x_edges: array-like (M,)
        Bin edges of the x-variable.
    y_edges: array-like (N,)
        Bin edges of the y-variable.
    dim: str or tuple
        Dimension name(s) over which to aggregate the histogram.
    **kwargs: optional
        Keyword arguments to be passed to `scipy.gaussian_kde()`

    Returns
    -------
    ret: xarray.DataArray
        DataArray containing the histogram with dimensions and
        coordinates set accordingly.

    See Also
    --------
    scipy.gaussian_kde
    """
    x_ctr = 0.5 * (x_edges[:-1] + x_edges[1:])
    y_ctr = 0.5 * (y_edges[:-1] + y_edges[1:])
    x_binv = x + "_bins"
    y_binv = y + "_bins"

    pX, pY = np.meshgrid(x_ctr, y_ctr)
    poss = np.vstack([pX.ravel(), pY.ravel()])

    gkde = gaussian_kde(
        ds[[x, y]].stack(_id=dims).dropna("_id").to_array(),
        **kwargs
    )
    info("kde bandwidth factor: %f", gkde.factor)
    kde_eval = gkde(poss)

    ret = xr.DataArray(
        kde_eval.reshape(pX.shape).T,
        dims=(x_binv, y_binv,),
        coords={x_binv: x_ctr, y_binv: y_ctr}
    )
    return ret


# %%
def xy_interpolate_to_zero(ds, bin_var, **kwargs):
    """Interpolate the mean, mode, or median NOy to zero CH4

    Interpolates the determined NOy per CH4 to zero CH4 vmr
    for interpolation and to calculate the background NOy from it.
    """
    _ds = ds.sel({bin_var: slice(-1e-12, None)})
    mask = np.isfinite(_ds)
    mi = np.where(mask)

    if np.all(_ds[mi].coords[bin_var] > 0.):
        # Add zero CH4 point for interpolation
        _zero_pt = xr.zeros_like(
            _ds.isel({bin_var: [0]})
        ).assign_coords({bin_var: [0.]})
        _ds = xr.concat([_zero_pt, _ds], dim=bin_var)
    # interpolate (over nans)
    ret = _ds.interpolate_na(dim=bin_var, **kwargs)
    return ret


# %%
def hist_mean(hist_da, x_var, y_var, min_hpts=50):
    _h_mean = (hist_da[y_var].data * hist_da).sum(y_var) / hist_da.sum(y_var)
    _hist_mean = xr.where(hist_da.sum(y_var) >= min_hpts, _h_mean, np.nan)
    debug("mean: %s", _hist_mean)
    return xy_interpolate_to_zero(_hist_mean, x_var, method="linear")


# %%
def hist_median(hist_da, x_var, y_var, min_hpts=50):
    _hcsum = hist_da.cumsum(y_var)
    _ix = []
    for _i in range(_hcsum.shape[0]):
        _ix.append(np.searchsorted(_hcsum[_i], 0.5 * hist_da.sum(y_var)[_i]))
    _h_median = hist_da.isel({y_var: _ix})[y_var]
    _h_median[x_var] = (y_var, hist_da[x_var].data)
    _h_median = _h_median.swap_dims({y_var: x_var})

    _hist_median = xr.where(hist_da.sum(y_var) >= min_hpts, _h_median, np.nan)
    debug("median: %s", _hist_median)
    return xy_interpolate_to_zero(_hist_median, x_var, method="linear")


# %%
def hist_mode(hist_da, x_var, y_var, min_hpts=50):
    _h_mode = hist_da.isel({y_var: hist_da.argmax(y_var)})[y_var]
    _hist_mode = xr.where(hist_da.sum(y_var) >= min_hpts, _h_mode, np.nan)
    debug("mode: %s", _hist_mode)
    return xy_interpolate_to_zero(_hist_mode, x_var, method="linear")



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
from os import path, sep as dsep

import pytest

from mipas_noxy.util import get_nc_filename, open_mipas_l2, open_mfmipas_l2


@pytest.mark.parametrize(
    "res, spec, v, name",
    [
        ["R", "NO", "261", "V8R_NO_261" + dsep + "MIPAS-E_IMK.200001.V8R_NO_261.nc"],
        ["H", "NO2", "61", "V8H_NO2_61" + dsep + "MIPAS-E_IMK.200001.V8H_NO2_61.nc"],
    ]
)
def test_get_nc_filename(res, spec, v, name):
    fpath = get_nc_filename(
        "", res, spec, 2000, 1, version=v,
    )
    assert fpath == name


def test_read_single():
    file = path.join(
        "tests", "data", "V8R_NO2_261_0", "MIPAS-E_IMK.201004.V8R_NO2_261_0.nc",
    )
    ds = open_mipas_l2(file)
    assert ds


@pytest.mark.parametrize(
    "files",
    [
        # glob version
        path.join(
            "tests", "data", "V8R_NO_261_0", "MIPAS-E_IMK.*.V8R_NO_261_0.nc",
        ),
        # list of paths
        [
            path.join(
                "tests", "data", "V8R_NO_261_0",
                "MIPAS-E_IMK.201004.V8R_NO_261_0.nc",
            )
        ],
    ]
)
def test_read_multi(files):
    ds = open_mfmipas_l2(files)
    assert ds



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2023 Stefan Bender
#
# This module is part of pymipas_noxy.
# pymipas_noxy is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying LICENSE file or http://www.gnu.org/licenses/gpl-2.0.html.
"""MIPAS NOxy calculator module
"""

__version__ = "0.0.1"

from .util import *



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2023 Stefan Bender
#
# This module is part of pymipas_noxy.
# pymipas_noxy is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying LICENSE file or http://www.gnu.org/licenses/gpl-2.0.html.
"""MIPAS NOxy calculator command line interface
"""
import logging

import numpy as np
import pandas as pd
import xarray as xr

import click
import toml

from .util import *

logging.basicConfig(
    level=logging.INFO,
    format=
        "[%(levelname)-8s] (%(asctime)s) "
        "%(filename)s:%(lineno)d:%(funcName)s() %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S %z",
)


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], show_default=True)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--config-file",
    default=None,
    help="Configuration file to use, takes precedence over the argument.",
)
@click.argument("configfile", required=False)
@click.option(
    "-m", "--month", default=None, type=int, help="month to process, needs also `year`."
)
@click.option(
    "-y", "--year", default=None, type=int, help="year to process, needs also `month`."
)
# Do not write output file
@click.option(
    "-n", "--dry-run", default=False, is_flag=True, help="Dry run, no output written."
)
# logging options
@click.option(
    "-v", "--verbose", "loglevel", flag_value="INFO", help="Sets 'loglevel' to 'INFO'."
)
@click.option(
    "-q", "--quiet", "loglevel", flag_value="ERROR", help="Sets 'loglevel' to 'ERROR'."
)
@click.option(
    "-l",
    "--loglevel",
    default="WARN",
    help="Sets 'loglevel' directly to the given level.",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR"], case_sensitive=False),
)
def main(
    config_file,
    configfile,
    month,
    year,
    dry_run,
    loglevel,
):
    try:
        xr.set_options(keep_attrs=True)
        xr.set_options(display_max_rows=1000, display_width=96)
    except ValueError:
        pass
    np.set_printoptions(precision=3)

    logger = logging.getLogger()
    debug = logger.debug
    info = logger.info
    warn = logger.warning
    error = logger.error
    info("setting loglevel: %s", loglevel)
    logger.setLevel(getattr(logging, loglevel.upper(), None))
    warn("new loglevel: %s", logging.getLogger().getEffectiveLevel())

    if not config_file:
        if not configfile:
            error(
                "No config file specified, "
                "use `--config-file` or pass as argument."
            )
            return -1
        else:
            config_file = configfile

    with open(config_file, mode="rb") as _fp:
        config = toml.loads(_fp.read().decode())
    debug("config: %s", config)

    # Time setup, command line takes precedence,
    # but works only if both year and month are given.
    if year and month:
        tymss = [(year, month)]
    else:
        times = config.get("times", [])
        tyms = [
            set((_t.year, _t.month) for _t in pd.date_range(**t))
            for t in times
        ]
        if not tyms:
            error(
                "No times specified, either on the cmmand line "
                "via `--year` and `--month`, or by using "
                "`[[times]]` entries in the configuration file."
            )
            return -2
        else:
            tymss = sorted(set.union(*tyms))
    info("year(s)/month(s): %s", tymss)

    # Short-cut config dictionaries
    inp_conf = config.get("input", {})
    out_conf = config.get("output", {})
    out_targets = out_conf.get("targets")
    if not hasattr(out_targets, "__getitem__"):
        out_targets = [out_targets]

    debug("input config: %s", inp_conf)
    debug("output config: %s", out_conf)
    info("output targets: %s", out_targets)

    for (year, month) in tymss:
        for out_target in out_targets:
            info("Processing: %s %s %s", out_target, year, month)

            # read species netcdfs into list
            mv8_noy_l = read_mv8_species_v1(config, out_target, year, month)
            if not mv8_noy_l:
                continue
            # interpolate to first (NO) altitudes
            mv8_noy_l = [mv8_noy_l[0]] + [
                interp_altitude(_d, altitude=mv8_noy_l[0].altitude, method="linear")
                for _d in mv8_noy_l[1:]
            ]
            debug("MIPAS v8 input list: %s", mv8_noy_l)

            # combine into one data set
            mv8_noy_ds = combine_NOxy(mv8_noy_l)
            debug("MIPAS v8 combined ds: %s", mv8_noy_ds)

            # sum/average to NOx/NOy
            mv8_noy = calculate_NOxy(mv8_noy_ds)
            debug("MIPAS v8 NOx/NOy ds: %s", mv8_noy)

            # drop unneeded variables
            mv8_noy1 = compat_drop_vars(
                mv8_noy,
                ["weights", "retrieval_in_logarithmic_parameter_space"]
            )

            # Fixup to make the dataset compliant with
            # the original v8 netcdf data sets.
            mv8_noy1 = fixup_target_name(
                mv8_noy1,
                inp_conf[out_target]["targets"][0],
                out_target.upper(),
            )
            mv8_noy1 = fixup_altitudes(mv8_noy1)
            info("MIPAS v8 NOx/NOy ds fixed: %s", mv8_noy1)

            # Construct file/path name for output species
            out_path = get_nc_filename(
                out_conf.get("path", "."),
                config.get("resolution", "R"),
                out_target.upper(),  # upper case in filename
                year,
                month,
                version=inp_conf.get(out_target, {}).get("versions", [])[0],
            )
            info("Saving to: %s", out_path)

            if dry_run:
                info("Dry run, writing skipped.")
            else:
                mv8_noy1.to_netcdf(out_path)
                info("Saved.")

    return 0


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
from subprocess import Popen

import pytest


@pytest.mark.parametrize(
    "binary",
    [
        "mipas_noxy",
        "python -m mipas_noxy",
    ],
)
def test_cli_help(binary):
    p = Popen(
        binary.split(" ") + [
            "-h",
        ],
    )
    p.communicate()
    p.wait()
    assert p.returncode == 0


@pytest.mark.parametrize(
    "binary",
    [
        "mipas_noxy",
        "python -m mipas_noxy",
    ],
)
def test_cli_dry_run(binary):
    p = Popen(
        binary.split(" ") + [
            "--config-file",
            "tests/test_NOx_config.toml",
            "-q",
            "-n",
        ],
    )
    p.communicate()
    p.wait()
    assert p.returncode == 0



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
import numpy as np
import xarray as xr

from mipas_noxy.util import combine_NOxy, calculate_NOxy, fixup_target_name, select_common_data

GEO_IDS = np.array(
    [b"00001_20000101T000001Z", b"00002_20000101T000101Z", b"00003_20000101T000201Z"]
)
TIMES = np.array(
    ["2000-01-01T00:00:01", "2000-01-01T00:01:01", "2000-01-01T00:02:01"],
    dtype="M8[ns]",
)


def _prep_dss():
    coords = {"time": TIMES, "altitude": [50.]}
    target_attrs = dict(
        standard_name = "mole_fraction_of_nitrogen_monoxide_in_air",
        units = "1e-6",
        long_name = "volume mixing ratio of NO",
    )
    global_attrs = dict(
        retrieval_target_name = "NO",
        retrieval_in_logarithmic_parameter_space = "TRUE",
    )
    ds1 = xr.Dataset(
        data_vars={
            "geo_id": (["time"], GEO_IDS),
            "dof": (["time"], [30, 30, 30]),
            "chi2": (["time"], [1.1, 1.1, 1.1]),
            "target": (["altitude", "time"], [[1., 1.5, 2.]], target_attrs),
            "target_noise_error": (["altitude", "time"], [[1., 1., 1.]]),
            "akm_diagonal": (["altitude", "time"], [[0.1, 0.15, 0.2]]),
            "weights": ([], 1.0),
        },
        coords=coords,
        attrs=global_attrs,
    ).expand_dims("species").assign_coords(species=["NO"])
    ds2 = xr.Dataset(
        data_vars={
            "geo_id": (["time"], GEO_IDS[[0, 2]]),
            "dof": (["time"], [40, 40]),
            "chi2": (["time"], [1.2, 1.2]),
            "target": (["altitude", "time"], [[3., 4.]], target_attrs),
            "target_noise_error": (["altitude", "time"], [[2., 2.]]),
            "akm_diagonal": (["altitude", "time"], [[0.2, 0.3]]),
            "weights": ([], 1.0),
        },
        coords = {"time": TIMES[[0, 2]], "altitude": [50.]},
        attrs=global_attrs,
    ).expand_dims("species").assign_coords(species=["NO2"])
    ds3 = xr.Dataset(
        data_vars={
            "geo_id": (["time"], GEO_IDS),
            "dof": (["time"], [50, 50, 50]),
            "chi2": (["time"], [1.3, 1.3, 1.3]),
            "target": (["altitude", "time"], [[-2., -1.5, -1.]], target_attrs),
            "target_noise_error": (["altitude", "time"], [[1., 1.5, 2.]]),
            "akm_diagonal": (["altitude", "time"], [[0.1, 0.15, 0.2]]),
            "weights": ([], 2.0),
        },
        coords=coords,
        attrs=global_attrs,
    ).expand_dims("species").assign_coords(species=["N2O5"])
    return [ds1, ds2, ds3]


def test_noxy():
    dss = _prep_dss()
    dsc = combine_NOxy(dss)
    noxy = calculate_NOxy(dsc)
    np.testing.assert_allclose(noxy.dof, [120, 120])
    np.testing.assert_allclose(noxy.target, [[0., 4.]])
    np.testing.assert_allclose(noxy.target_noise_error, [[3., 4.582576]])
    np.testing.assert_allclose(noxy.chi2, [1.216666667, 1.216666667])
    np.testing.assert_allclose(noxy.akm_diagonal, [[.1375, 0.25]])
    assert (noxy.geo_id == GEO_IDS[[0, 2]]).all()
    assert (noxy.time == TIMES[[0, 2]]).all()


def test_fixup_target_name():
    noxy = _prep_dss()[0]
    noxy = fixup_target_name(noxy, "NO", "NOY")
    assert noxy.target.attrs["long_name"] == "volume mixing ratio of NOY"
    assert noxy.target.attrs["standard_name"] == "mole_fraction_of_noy_expressed_as_nitrogen_in_air"
    assert noxy.attrs["retrieval_target_name"] == "NOY"


def test_select_common():
    dss = _prep_dss()
    dss = [ds.isel(species=0) for ds in dss]
    sel = (select_common_data(dss, on="geo_id"))
    for ds in sel:
        np.testing.assert_equal(
            ds["geo_id"].values,
            [b"00001_20000101T000001Z", b"00003_20000101T000201Z"],
        )
        np.testing.assert_equal(
            ds["time"].values,
            np.array(["2000-01-01T00:00:01", "2000-01-01T00:02:01"], dtype="M8"),
        )



from codecs import open
from os import path
import re

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

name = "mipas_noxy"
meta_path = path.join("src", name, "__init__.py")
here = path.abspath(path.dirname(__file__))

extras_require = {
    "tests": ["pytest"],
}
extras_require["all"] = sorted(
    {v for req in extras_require.values() for v in req}
)


# Approach taken from
# https://packaging.python.org/guides/single-sourcing-package-version/
# and the `attrs` package https://www.attrs.org/
# https://github.com/python-attrs/attrs
def read(*parts):
    """
    Builds an absolute path from *parts* and and return the contents of the
    resulting file.  Assumes UTF-8 encoding.
    """
    with open(path.join(here, *parts), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta, *path):
    """
    Extract __*meta*__ from *path* (can have multiple components)
    """
    meta_file = read(*path)
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        meta_file,
        re.M,
    )
    if not meta_match:
        raise RuntimeError("__{meta}__ string not found.".format(meta=meta))
    return meta_match.group(1)


# Get the long description from the README file
long_description = read("README.md")
version = find_meta("version", meta_path)

if __name__ == "__main__":
    setup(
        name=name,
        version=version,
        description="MIPAS level-2 NOx/NOy calculater",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/st-bender/pymipas_noxy",
        author="Stefan Bender",
        author_email="sbender@iaa.es",
        license="GPLv2",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Utilities",
        ],
        packages=find_packages("src"),
        package_dir={"": "src"},
        include_package_data=True,
        install_requires=[
            "numpy",
            "scipy",
            "xarray",
            "h5netcdf",
            "click",
            "toml",
        ],
        extras_require=extras_require,
        scripts=[],
        entry_points={
            "console_scripts": [
                "mipas_noxy = mipas_noxy.__main__:main",
            ]
        },
        options={
            "bdist_wheel": {"universal": True},
        },
        zip_safe=False,
    )



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2023 Stefan Bender
#
# This module is part of pymipas_noxy.
# pymipas_noxy is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying LICENSE file or http://www.gnu.org/licenses/gpl-2.0.html.
"""MIPAS NOx and NOy calculator utility functions

Utility functions to read MIPAS level-2 netcdf files and
to calculate NOx or NOy species from the individual species.
"""
from __future__ import absolute_import, division, print_function

from logging import info, debug
from os import path

import h5netcdf.legacyapi as nc4
import numpy as np
import xarray as xr

from scipy.io import netcdf_file


def get_nc_filename(l2_path, resolution, species, year, month, ext="nc", version="61"):
    """Infer the full path for the level-2 netcdf files
    """
    species_subdir = "V8{resolution}_{species}_{version}".format(
        resolution=resolution,
        species=species,
        version=version,
    )
    file = "MIPAS-E_IMK.{year:04d}{month:02d}.V8{resolution}_{species}_{version}.{ext}".format(
        year=year,
        month=month,
        resolution=resolution,
        species=species,
        version=version,
        ext=ext,
    )
    full_path = path.join(
        l2_path,
        species_subdir,
        file
    )
    return full_path


def open_mipas_l2(file, **kwargs):
    """Open and harmonize MIPAS level-2 netcdf file

    Opens the MIPAS level-2 netcdf file and harmonizes the
    altitude coordinates so that the dataset can be easily
    processed with `xarray`.

    Parameters
    ----------
    file: str
        The name of the file to open.
    **kwargs: dict, optional
        Keyword arguments passed to `xarray.open_dataset()`

    Returns
    -------
    dataset: xarray.Dataset
        The dataset containing the MIPAS level-2 data.
    """
    def _open_nc34(file):
        try:
            # try netcdf4 with h5netcdf first
            _ds = nc4.Dataset(file, mode="r")
        except (IOError, OSError):
            # apparently a netcdf3 file, use scipy.io
            _ds = netcdf_file(file, mmap=False)
        return _ds

    def _get_attrs(v):
        if hasattr(v, "attrs"):
            # h5netcdf netcdf4
            attrs = v.attrs
        else:
            # netCDF4 just in case
            attrs = v.__dict__
            if "_attributes" in attrs:
                # scipy.io netcdf3
                attrs = attrs["_attributes"]
        # at least scipy.io converts the attribute values to bytes,
        # converts them back to be consistent with the others.
        attrs.update({
            k: _v.decode() for k, _v in attrs.items()
            if type(_v) == bytes
        })
        return attrs

    # Drops the "altitude" variable since `xarray` can't handle
    # dimensions and variables with the same name but different dimensions
    mipv8_ds = xr.open_dataset(file, drop_variables=["altitude"], **kwargs)
    if "altgrid" in mipv8_ds.dims:
        mipv8_ds = mipv8_ds.rename({"altgrid": "altitude"})
    if "timegrid" in mipv8_ds.dims:
        mipv8_ds = mipv8_ds.swap_dims({"timegrid": "time"})
    # Round time to 1 millisecond.
    # Typically the `time` variable comes as double precision
    # days since .... which is only accurate to a few 10s µs.
    # The `geo_id`s are based on seconds, so milliseconds should be good enough.
    mipv8_ds["time"] = mipv8_ds.time.dt.round("1ms")
    # harmonize dtypes to avoid big/little endian mixup
    _dtype = mipv8_ds.target.dtype
    # Fix altitude naming and coordinates
    nc4_ds = _open_nc34(file)

    # Checks if `altitude` has more than one dimension
    # and copies that to a separate variable to keep
    # `altitude` as 1-d coordinates for further processing.
    alt_var = nc4_ds.variables["altitude"]
    alt_attrs = _get_attrs(alt_var)
    alt_values = alt_var[:].astype(_dtype)
    if len(alt_var.dimensions) > 1:
        _alt = xr.DataArray(
            alt_values,
            dims=("altitude", "time"),
            coords={
                "altitude": ("altitude", alt_values[:, 0], alt_attrs),
                "time": mipv8_ds.time
            },
            attrs=alt_attrs,
        )
        # Re-creates the original "altitude" variable but with a different name
        mipv8_ds["alt"] = _alt
    else:
        mipv8_ds["altitude"] = ("altitude", alt_values, alt_attrs)
    return mipv8_ds


def open_mfmipas_l2(files, dim="time", **kwargs):
    """Opens multiple MIPAS level-2 files

    Automatically combines multiple MIPAS level-2 netcdf data files
    along the concatenation (time) dimension. Uses `open_mipas_l2`
    to iterate over the files passed or derived from the glob string.
    Currently works only for opening files for the *same* trace gas.

    Parameters
    ----------
    files: str or sequence of paths
        The file names to open, either as a glob string or a list
        of path names.
    dim: str, optional (default: time)
        Dimension over which to concatenate the level-2 datasets.
    **kwargs: dict, optional
        Keyword arguments passed to `xarray.open_dataset()`

    Returns
    -------
    dataset: xarray.Dataset
        The dataset containing the MIPAS level-2 data.
    """
    def _fspath1(f):
        # mimic `os.fspath()` for Py27 and Py35
        # no error checking
        return type(f).__fspath__(f)

    # inspired by `xarray.backends.common._find_absolute_paths()`
    if isinstance(files, str):
        from glob import glob
        files = sorted(glob(files))
    elif hasattr(files, "__fspath__"):
        files = [_fspath1(files)]
    else:
        files = [_fspath1(f) if hasattr(f, "__fspath__") else f for f in files]

    datasets = [open_mipas_l2(f, **kwargs) for f in files]
    try:
        combined = xr.concat(datasets, dim=dim)
    except (TypeError, ValueError):
        # close datasets on errors
        for ds in datasets:
            ds.close()
        raise

    return combined


def interp_altitude(ds, interp_error=False, **kwargs):
    ret = ds.interp(**kwargs)
    if interp_error:
        # overwrite `target_noise_error` with the squared interpolation
        ret["target_noise_error"] = np.sqrt(
            (ds.target_noise_error**2).interp(**kwargs)
        )
    return ret


def _read_mipas_species(file, _s, _w, interp_alts=None, load_kwargs=None):
    debug("%s %s", file, path.exists(file))
    with open_mipas_l2(file, **load_kwargs) as _ds:
        _ds = _ds.expand_dims("species").assign_coords(species=[_s])
        _ds["weights"] = ("species", [_w])
        log_retr = (
            _ds.attrs["retrieval_in_logarithmic_parameter_space"].lower()
            in ["1", "true", "yes"]
        )
        _ds["retrieval_in_logarithmic_parameter_space"] = ("species", [log_retr])
    return _ds


def read_mv8_species_v1(config, target, year, month, load_kwargs=None):
    load_kwargs = load_kwargs or {}
    _res = config.get("resolution", "R")

    input_conf = config.get("input", {})
    input_path = input_conf.get("path")

    spec_conf = input_conf[target]
    targets = spec_conf.get("targets", [])
    versions = spec_conf.get("versions", [])
    weights = spec_conf.get("weights", [])

    # iterate over the targets
    dsl = []
    for _s, _v, _w in zip(targets, versions, weights):
        info("species: %s, version: %s, weight: %s", _s, _v, _w)
        _file = get_nc_filename(input_path, _res, _s, year, month, version=_v)
        if not path.exists(_file):
            return None
        _ds = _read_mipas_species(_file, _s, _w, load_kwargs=load_kwargs)
        dsl.append(_ds)
    return dsl


def combine_NOxy(dsl, species=None):
    # order "species" as in the list, or as given as an argument
    if species is None:
        species = [_d.species.values[0] for _d in dsl]
    _ds = xr.merge(dsl)
    _ds = _ds.sel(species=species)
    _dsb = _ds.dropna(dim="time", how="any", subset=["geo_id"])

    geo_id_set = set(_ds.geo_id.values[0])
    for gids in _ds.geo_id.values[1:]:
        geo_id_set = geo_id_set.intersection(set(gids))
    # Filters out NaNs by type "float" since using `np.nan` didn't work.
    # Ensures that the `byte` list can be sorted and compared
    # to assert that we have indeed all the available geo_ids
    # across all data sets in the combined data set.
    geo_id_set = list(filter(lambda o: type(o) != float, geo_id_set))
    assert (sorted(geo_id_set) == _dsb.geo_id).all()
    # fix dtypes
    for _v in dsl[0].data_vars:
        _dsb[_v] = _dsb[_v].astype(dsl[0][_v].dtype)
    return _dsb


def calculate_NOxy(ds, dim="species", keep_attrs=True):
    scaled_vars = ["target"]
    squared_vars = ["target_noise_error"]
    dof_vars = ["chi2", "rms"]
    akm_vars = ["akm_diagonal", "vr_akdiag", "vr_col", "vr_row"]

    _dsvars = set(ds.data_vars)

    # copy first dataset and update later
    mv8_noxy = ds.isel({dim: 0}).drop(dim).copy()

    # overwrite degrees of freedom with the sum
    mv8_noxy["dof"] = ds.dof.sum(dim=dim)

    # update scaled (weighted) variables with the weighted (scaled) sum
    # usually: weights = number of N (nitrogen) atoms in molecule
    for _v in _dsvars & set(scaled_vars):
        mv8_noxy[_v] = (ds[_v] * ds.weights).sum(dim=dim)

    # update "squared" variables with the sqrt(sum of squares)
    # e.g. error variables
    for _v in _dsvars & set(squared_vars):
        mv8_noxy[_v] = np.sqrt(((ds[_v] * ds.weights)**2).sum(dim=dim))

    # update dof-weighted variables with the dof-weighted average
    for _v in _dsvars & set(dof_vars):
        mv8_noxy[_v] = (ds[_v] * ds.dof).sum(dim=dim) / mv8_noxy.dof

    # update akm-like variables with the target-weighted average
    ak_ww = ds.weights * np.abs(ds.target)
    ak_w = ak_ww / ak_ww.sum(dim=dim)
    for _v in _dsvars & set(akm_vars):
        mv8_noxy[_v] = (ds[_v] * ak_w).sum(dim=dim)

    for _v in mv8_noxy.data_vars:
        # fix dtype
        mv8_noxy[_v] = mv8_noxy[_v].astype(ds[_v].dtype)
        if keep_attrs:
            # copy attributes
            mv8_noxy[_v].attrs = ds[_v].attrs

    return mv8_noxy


# from v81, 25 Apr 2023 of
# https://cfconventions.org/Data/cf-standard-names/current/build/cf-standard-name-table.html
STD_NAMES = {
	"NOX": "mole_fraction_of_nox_expressed_as_nitrogen_in_air",
	"NOY": "mole_fraction_of_noy_expressed_as_nitrogen_in_air",
}


def fixup_target_name(ds, from_name, to_name):
    # We are going to overwrite things, so make a copy first.
    ds = ds.copy()
    # replace in variable attributes
    for _v in ds.data_vars:
        _attrs = ds[_v].attrs
        _ln = _attrs.get("long_name", None)
        _sn = _attrs.get("standard_name", None)
        if _ln and from_name in _ln:
            _attrs["long_name"] = _ln.replace(from_name, to_name)
        if _sn and _sn.startswith("mole_fraction"):
            _attrs["standard_name"] = STD_NAMES.get(to_name.upper(), _sn)
    # replace in global attributes
    for _k, _v in ds.attrs.items():
        if from_name in _v:
            ds.attrs[_k] = _v.replace(from_name, to_name)
    return ds


def compat_drop_vars(ds, var):
    """Backwards-compatible dropping of vairbales in the dataset

    Just in case one or the other does not exist or gets removed
    from the API.
    """
    try:
        ds = ds.drop_vars(var)
    except AttributeError:
        ds = ds.drop(var)
    return ds


def fixup_altitudes(ds, alt_name="alt"):
    if alt_name not in ds.data_vars:
        info(
            "Variable '%s' not found, skip resetting altitude coordinates",
            alt_name
        )
        return ds
    # drops the "altitude" coordinate variable first
    # if there is already one in the dataset
    if "altitude" in ds.coords:
        ds = compat_drop_vars(ds, ["altitude"])
    # the variable in `alt_name` should hold the
    # "original" time-dependent altitude coordinates,
    # rename it to be consistent with the input nc files.
    ds = ds.rename({alt_name: "altitude"})
    return ds


def select_common_data(dss, dim="time", on="geo_id"):
    """Select common IDs in datasets

    Selects the datasets according to their common (unique) IDs,
    e.g. the `geo_id` time stamp for MIPAS V8.

    Parameters
    ----------
    dss: list of `xarray.Dataset`
        List of datasets to be combined, based on their common geo_ids.
    dim: str, optional (default: time)
        Dimension along which the ID variable is to be selected.
    on: str, optional (default: geo_id)
        Variable that contains the (unique) IDs in each dataset.

    Returns
    -------
    ret: list of `xarray.Dataset`
        List of the datasets, trimmed to the common (unique) IDs.
    """
    _ids = sorted(set.intersection(*map(set, map(lambda d: d[on].values, dss))))
    ret = [d.swap_dims({dim: on}).sel({on: _ids}).swap_dims({on: dim}) for d in dss]
    return ret
